"""Tests for deprecation of old, non-PEP8 API.

These tests can be removed when the deprecated classes/methods/functions are
removed from CyIpopt.

"""


import pytest

import cyipopt


def test_ipopt_import_deprecation():
    """Ensure that old module name import raises FutureWarning to user."""
    expected_warning_msg = ("The module has been renamed to 'cyipopt' from "
                            "'ipopt'. Please import using 'import cyipopt' "
                            "and remove all uses of 'import ipopt' in your "
                            "code as this will be deprecated in a future "
                            "release.")
    with pytest.warns(FutureWarning, match=expected_warning_msg):
        import ipopt


def test_non_pep8_class_name_deprecation(hs071_defintion_instance_fixture,
                                         hs071_initial_guess_fixture,
                                         hs071_variable_lower_bounds_fixture,
                                         hs071_variable_upper_bounds_fixture,
                                         hs071_constraint_lower_bounds_fixture,
                                         hs071_constraint_upper_bounds_fixture,
                                         ):
    """Ensure use of old non-PEP8 classes API raises FutureWarning to user."""
    expected_warning_msg = ("The class named 'problem' will soon be "
                            "deprecated in CyIpopt. Please replace all uses "
                            "and use 'Problem' going forward.")
    with pytest.warns(FutureWarning, match=expected_warning_msg):
        _ = cyipopt.problem(n=len(hs071_initial_guess_fixture),
                            m=len(hs071_constraint_lower_bounds_fixture),
                            problem_obj=hs071_defintion_instance_fixture,
                            lb=hs071_variable_lower_bounds_fixture,
                            ub=hs071_variable_upper_bounds_fixture,
                            cl=hs071_constraint_lower_bounds_fixture,
                            cu=hs071_constraint_upper_bounds_fixture,
                            )


def test_non_pep8_set_logging_level_deprecation():
    """Ensure use of old non-PEP8 classes API raises FutureWarning to user."""
    expected_warning_msg = ("The function named 'setLoggingLevel' will soon "
                            "be deprecated in CyIpopt. Please replace all "
                            "uses and use 'set_logging_level' going forward.")
    with pytest.warns(FutureWarning, match=expected_warning_msg):
        cyipopt.setLoggingLevel()


def test_non_pep8_method_names_deprecation(hs071_problem_instance_fixture):
    """Ensure use of old non-PEP8 methods API raises FutureWarning to user."""
    nlp = hs071_problem_instance_fixture

    assert isinstance(nlp, cyipopt.Problem)

    expected_warning_msg = ""
    with pytest.warns(FutureWarning, match=expected_warning_msg):
        nlp.addOption("mu_strategy", "adaptive")

    expected_warning_msg = ""
    with pytest.warns(FutureWarning, match=expected_warning_msg):
        nlp.setProblemScaling(obj_scaling=2.0)


def test_deprecated_problem_can_be_subclassed():
    """`problem` can be subclassed and its args/kwargs changed."""

    class SubclassedProblem(cyipopt.problem):

        def __init__(self, *args, **kwargs):
            n = args[0]
            m = args[1]
            problem_obj = kwargs.get("problem_obj")
            lb = kwargs.get("lb")
            ub = kwargs.get("ub")
            cl = kwargs.get("cl")
            cu = kwargs.get("cu")
            super(SubclassedProblem, self).__init__(n,
                                                    m,
                                                    problem_obj=problem_obj,
                                                    lb=lb,
                                                    ub=ub,
                                                    cl=cl,
                                                    cu=cu)

        def objective(self):
            pass

        def gradient(self):
            pass

        def constraints(self):
            pass

        def jacobian(self):
            pass

    expected_warning_msg = ("The class named 'problem' will soon be "
                            "deprecated in CyIpopt. Please replace all uses "
                            "and use 'Problem' going forward.")
    with pytest.warns(FutureWarning, match=expected_warning_msg):
        _ = SubclassedProblem(2, 2, None, None, problem_obj=None, lb=None,
                              ub=None, cl=[0, 0], cu=[0, 0], other_kwarg=None)



"""Fixture-sharing file for test suite.

This file includes a default instantiation of the hs071 test problem used
elsewhere within this package for demonstration purposes.
"""

import builtins
import sys

import numpy as np
import pytest

import cyipopt


@pytest.fixture()
def hs071_objective_fixture():
    """Return a function for the hs071 test problem objective."""

    def objective(x):
        return x[0] * x[3] * np.sum(x[0:3]) + x[2]

    return objective


@pytest.fixture()
def hs071_gradient_fixture():
    """Return a function for the hs071 test problem gradient."""

    def gradient(x):
        return np.array([x[0] * x[3] + x[3] * np.sum(x[0:3]),
                         x[0] * x[3],
                         x[0] * x[3] + 1.0,
                         x[0] * np.sum(x[0:3])
                         ])

    return gradient


@pytest.fixture()
def hs071_constraints_fixture():
    """Return a function for the hs071 test problem constraints."""

    def constraints(x):
        return np.array((np.prod(x), np.dot(x, x)))

    return constraints


@pytest.fixture()
def hs071_jacobian_fixture():
    """Return a function for the hs071 test problem jacobian."""

    def jacobian(x):
        return np.concatenate((np.prod(x) / x, 2 * x))

    return jacobian


@pytest.fixture()
def hs071_jacobian_structure_fixture():
    """Return a function for the hs071 test problem jacobian structure."""

    def jacobian_structure(x):
        return np.ones(1, 4)

    return jacobian_structure


@pytest.fixture()
def hs071_hessian_fixture():
    """Return a function for the hs071 test problem hessian."""

    def hessian(x, lagrange, obj_factor):
        H = obj_factor * np.array(((2 * x[3], 0, 0, 0),
                                   (x[3], 0, 0, 0),
                                   (x[3], 0, 0, 0),
                                   (2 * x[0] + x[1] + x[2], x[0], x[0], 0)
                                   ))
        H += lagrange[0] * np.array(((0, 0, 0, 0),
                                     (x[2] * x[3], 0, 0, 0),
                                     (x[1] * x[3], x[0] * x[3], 0, 0),
                                     (x[1] * x[2], x[0] * x[2], x[0] * x[1], 0)
                                     ))
        H += lagrange[1] * 2 * np.eye(4)
        row, col = np.nonzero(np.tril(np.ones((4, 4))))
        return H[row, col]

    return hessian


@pytest.fixture()
def hs071_hessian_structure_fixture():
    """Return a function for the hs071 test problem hessian structure."""

    def hessian_structure(x):
        return np.nonzero(np.tril(np.ones((4, 4))))

    return hessian_structure


@pytest.fixture()
def hs071_intermediate_fixture():
    """Return a function for a default intermediate function."""
    def intermediate(*args):
        iter_count = args[2]
        obj_value = args[3]
        msg = f"Objective value at iteration #{iter_count} is - {obj_value}"
        print(msg)

    return intermediate


@pytest.fixture()
def hs071_defintion_instance_fixture(hs071_objective_fixture,
                                     hs071_gradient_fixture,
                                     hs071_constraints_fixture,
                                     hs071_jacobian_fixture,
                                     hs071_jacobian_structure_fixture,
                                     hs071_hessian_fixture,
                                     hs071_hessian_structure_fixture,
                                     hs071_intermediate_fixture,
                                     ):
    """Return a default implementation of the hs071 test problem."""

    class hs071:
        """The hs071 test problem also found in examples."""

        def __init__(self):
            self.objective = hs071_objective_fixture
            self.gradient = hs071_gradient_fixture
            self.constraints = hs071_constraints_fixture
            self.jacobian = hs071_jacobian_fixture
            self.jacobian_structure = hs071_jacobian_structure_fixture
            self.hessian = hs071_hessian_fixture
            self.hessian_structure = hs071_hessian_structure_fixture
            self.intermediate = hs071_intermediate_fixture

    problem_instance = hs071()
    return problem_instance


@pytest.fixture()
def hs071_initial_guess_fixture():
    """Return a default initial guess for the hs071 test problem."""
    x0 = [1.0, 5.0, 5.0, 1.0]
    return x0


@pytest.fixture()
def hs071_variable_lower_bounds_fixture():
    """Return a default variable lower bounds for the hs071 test problem."""
    lb = [1.0, 1.0, 1.0, 1.0]
    return lb


@pytest.fixture()
def hs071_variable_upper_bounds_fixture():
    """Return a default variable upper bounds for the hs071 test problem."""
    ub = [5.0, 5.0, 5.0, 5.0]
    return ub


@pytest.fixture()
def hs071_constraint_lower_bounds_fixture():
    """Return a default constraint lower bounds for the hs071 test problem."""
    cl = [25.0, 40.0]
    return cl


@pytest.fixture()
def hs071_constraint_upper_bounds_fixture():
    """Return a default constraint upper bounds for the hs071 test problem."""
    cu = [2.0e19, 40.0]
    return cu


@pytest.fixture()
def hs071_problem_instance_fixture(hs071_defintion_instance_fixture,
                                   hs071_initial_guess_fixture,
                                   hs071_variable_lower_bounds_fixture,
                                   hs071_variable_upper_bounds_fixture,
                                   hs071_constraint_lower_bounds_fixture,
                                   hs071_constraint_upper_bounds_fixture,
                                   ):
    """Return a default cyipopt.Problem instance of the hs071 test problem."""
    problem_definition = hs071_defintion_instance_fixture
    x0 = hs071_initial_guess_fixture
    lb = hs071_variable_lower_bounds_fixture
    ub = hs071_variable_upper_bounds_fixture
    cl = hs071_constraint_lower_bounds_fixture
    cu = hs071_constraint_upper_bounds_fixture
    n = len(x0)
    m = len(cl)
    problem = cyipopt.Problem(n=n, m=m, problem_obj=problem_definition, lb=lb,
                              ub=ub, cl=cl, cu=cu)
    return problem


@pytest.fixture
def hs071_problem_instance_with_options_fixture(hs071_problem_instance_fixture):
    nlp = hs071_problem_instance_fixture
    nlp.add_option('mu_strategy', 'adaptive')
    nlp.add_option('tol', 1e-7)
    return nlp



"""Backward-compatible module."""

import warnings

from .scipy_interface import IpoptProblemWrapper as CyipoptIpoptProblemWrapper
from .scipy_interface import convert_to_bytes as cyipopt_convert_to_bytes
from .scipy_interface import get_bounds as cyipopt_get_bounds
from .scipy_interface import get_constraint_bounds as cyipopt_get_constraint_bounds
from .scipy_interface import minimize_ipopt as cyipopt_minimize_ipopt


__all__ = ["get_bounds", "minimize_ipopt"]


def make_future_warning_error_msg(func_name):
    msg = (f"'{func_name}' from 'ipopt.ipopt_wrapper' has been replaced by "
           f"'{func_name}' from 'cyipopt.scipy_interface'. Please import using "
           f"'from cyipopt.scipy_interface import {func_name}' (or similar) and "
           f"remove all references to 'ipopt.ipopt_wrapper.{func_name}' in "
           f"your code as this will be deprecated in a future release.")
    return msg


def convert_to_bytes(*args, **kwargs):
    """Wrapper around `convert_to_bytes` for backwards compatibility."""
    msg = make_future_warning_error_msg("convert_to_bytes")
    warnings.warn(msg, FutureWarning)
    return cyipopt_convert_to_bytes(*args, **kwargs)


def get_bounds(*args, **kwargs):
    """Wrapper around `get_bounds` for backwards compatibility."""
    msg = make_future_warning_error_msg("get_bounds")
    warnings.warn(msg, FutureWarning)
    return cyipopt_get_bounds(*args, **kwargs)


def get_constraint_bounds(*args, **kwargs):
    """Wrapper around `get_constraint_bounds` for backwards compatibility."""
    msg = make_future_warning_error_msg("get_constraint_bounds")
    warnings.warn(msg, FutureWarning)
    return cyipopt_get_constraint_bounds(*args, **kwargs)


def minimize_ipopt(*args, **kwargs):
    """Wrapper around `minimize_ipopt` for backwards compatibility."""
    msg = make_future_warning_error_msg("minimize_ipopt")
    warnings.warn(msg, FutureWarning)
    return cyipopt_minimize_ipopt(*args, **kwargs)


def replace_option(options, oldname, newname):
    """Wrapper around `replace_option` for backwards compatibility."""
    msg = make_future_warning_error_msg("replace_option")
    warnings.warn(msg, FutureWarning)
    return cyipopt_replace_option(*args, **kwargs)


class IpoptProblemWrapper:
    """Wrapper around `IpoptProblemWrapper` for backwards compatibility."""

    def __new__(self, *args, **kwargs):
        msg = make_future_warning_error_msg("IpoptProblemWrapper")
        warnings.warn(msg, FutureWarning)
        return CyipoptIpoptProblemWrapper(*args, **kwargs)
